// VIVector.h: interface of the CVIVector class
//
/*--------------------------------------------------------------------------*/
/*       Copyright(C) 2004-2008 KONICA MINOLTA SENSING,INC.                 */
/*--------------------------------------------------------------------------*/

#if !defined(__VIVECTOR_H__)
#define __VIVECTOR_H__

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include <math.h>

#include <assert.h>

#define FEQ_EPS2    1e-12

#ifdef DATAPROCESS_EXPORTS
#define DATAPROCESS_API __declspec(dllexport)
#else
#define DATAPROCESS_API __declspec(dllimport)
#endif


class DATAPROCESS_API CVIVectorND  
{
    double* v;

public:
    CVIVectorND(){ v = 0;}
    CVIVectorND(int n);
    ~CVIVectorND();
    int nOrder;
        
    double& operator()(int i);
    double  operator()(int i) const;

    CVIVectorND    operator + (const CVIVectorND& v);
    CVIVectorND    operator - (const CVIVectorND& v);
    CVIVectorND&   operator = (const CVIVectorND& v);
    double       operator * (const CVIVectorND& v);
    CVIVectorND    operator * (const double v);
    CVIVectorND    operator / (const double v);
    int         operator ==(const CVIVectorND& v);
    int         operator !=(const CVIVectorND& v);

    double& Value(int i);
    BOOL        SetVector(int n);
    void        Normalize();
    int&        Order();

    double  EMPTY;

};

///////////////////////////////////////////////////////////////////////////////
class DATAPROCESS_API CVIVector2
{
private:
    double m_d[2];
    
protected:
    inline void Copy(const CVIVector2& v);
    
public:
    //-----------------------------------------------------------------------
    inline CVIVector2()                     {                                   }
    inline CVIVector2(double x, double y)   { Set(x, y);                        }
    inline CVIVector2(const CVIVector2& v)  { Copy(v);                          }
    inline CVIVector2(const double *v)      { Set(v[0], v[1]);                  }
    
    //-----------------------------------------------------------------------
    inline double& operator[](int i)       { assert(i>=0 && i<2); return m_d[i]; }
    inline double  operator[](int i) const { assert(i>=0 && i<2); return m_d[i]; }
    
    inline double* Raw()                { return m_d;                       }
    inline const double* Raw() const    { return m_d;                       }
    
    //-----------------------------------------------------------------------
    inline BOOL operator==(const CVIVector2& v) const;
    inline BOOL operator!=(const CVIVector2& v) const;
    
    //-----------------------------------------------------------------------
    inline void Set(double x, double y);
    inline void Set(const double *v);
    inline CVIVector2& operator=(const CVIVector2& v);
    inline CVIVector2& operator+=(const CVIVector2& v);
    inline CVIVector2& operator-=(const CVIVector2& v);
    inline CVIVector2& operator*=(double s);
    inline CVIVector2& operator/=(double s);
    
    //-----------------------------------------------------------------------
    inline CVIVector2 operator+(const CVIVector2& v) const;
    inline CVIVector2 operator-(const CVIVector2& v) const;
    inline CVIVector2 operator-() const;
    inline CVIVector2 operator*(double s) const;
    inline CVIVector2 operator/(double s) const;
    inline double   operator*(const CVIVector2& v) const;
};

/////////////////////////////////////////////////////////////////////////////
/*-------------------------------------------------------------------------*/
inline DATAPROCESS_API double Norm(const CVIVector2& v)
{
    return sqrt(v[0]*v[0] + v[1]*v[1]);
}

/*-------------------------------------------------------------------------*/
inline DATAPROCESS_API double Norm2(const CVIVector2& v)
{
    return v[0]*v[0] + v[1]*v[1];
}

/*-------------------------------------------------------------------------*/
inline DATAPROCESS_API double Length(const CVIVector2& v)
{
    return Norm(v);
}

/*-------------------------------------------------------------------------*/
inline DATAPROCESS_API double Length2(const CVIVector2& v)
{
    return Norm2(v);
}

/*-------------------------------------------------------------------------*/
inline DATAPROCESS_API double Unitize(CVIVector2& v)
{
    double l=Norm2(v);
    if( l!=1.0 && l!=0.0 )
    {
        l = sqrt(l);
        v /= l;
    }
    return l;
}

/*-------------------------------------------------------------------------*/
inline CVIVector2 operator*(double s, const CVIVector2& v)
{
    return v*s;
}

/////////////////////////////////////////////////////////////////////////////
/*-------------------------------------------------------------------------*/
inline void CVIVector2::Copy(const CVIVector2& v)
{
    m_d[0] = v[0];  m_d[1] = v[1];
}

/*-------------------------------------------------------------------------*/
inline void CVIVector2::Set(double x, double y)
{
    m_d[0] = x;     m_d[1] = y;
}

/*-------------------------------------------------------------------------*/
inline void CVIVector2::Set(const double *v)
{
    m_d[0] = v[0];  m_d[1] = v[1];
}

/*-------------------------------------------------------------------------*/
inline BOOL CVIVector2::operator==(const CVIVector2& v) const
{
    double dx = m_d[0] - v[0];
    double dy = m_d[1] - v[1];
    return (dx*dx + dy*dy) <= FEQ_EPS2;
}

/*-------------------------------------------------------------------------*/
inline BOOL CVIVector2::operator!=(const CVIVector2& v) const
{
    double dx = m_d[0] - v[0];
    double dy = m_d[1] - v[1];
    return (dx*dx + dy*dy) > FEQ_EPS2;
}

/*-------------------------------------------------------------------------*/
inline CVIVector2& CVIVector2::operator=(const CVIVector2& v)
{
    Copy(v);
    return *this;
}

/*-------------------------------------------------------------------------*/
inline CVIVector2& CVIVector2::operator+=(const CVIVector2& v)
{
    m_d[0] += v[0]; m_d[1] += v[1];
    return *this;
}

/*-------------------------------------------------------------------------*/
inline CVIVector2& CVIVector2::operator-=(const CVIVector2& v)
{
    m_d[0] -= v[0]; m_d[1] -= v[1];
    return *this;
}

/*-------------------------------------------------------------------------*/
inline CVIVector2& CVIVector2::operator*=(double s)
{
    m_d[0] *= s;    m_d[1] *= s;
    return *this;
}

/*-------------------------------------------------------------------------*/
inline CVIVector2& CVIVector2::operator/=(double s)
{
    double t = 1/s;
    m_d[0] *= t;    m_d[1] *= t;
    return *this;
}

/*-------------------------------------------------------------------------*/
inline CVIVector2 CVIVector2::operator+(const CVIVector2& v) const
{
    return CVIVector2(m_d[0]+v[0], m_d[1]+v[1]);
}

/*-------------------------------------------------------------------------*/
inline CVIVector2 CVIVector2::operator-(const CVIVector2& v) const
{
    return CVIVector2(m_d[0]-v[0], m_d[1]-v[1]);
}

/*-------------------------------------------------------------------------*/
inline CVIVector2 CVIVector2::operator-() const
{
    return CVIVector2(-m_d[0], -m_d[1]);
}

/*-------------------------------------------------------------------------*/
inline CVIVector2 CVIVector2::operator*(double s) const
{
    return CVIVector2(m_d[0]*s, m_d[1]*s);
}

/*-------------------------------------------------------------------------*/
inline CVIVector2 CVIVector2::operator/(double s) const
{
    double t = 1/s;
    return CVIVector2(m_d[0]*t, m_d[1]*t);
}

/*-------------------------------------------------------------------------*/
inline double CVIVector2::operator*(const CVIVector2& v) const
{
    return m_d[0]*v[0] + m_d[1]*v[1];
}


///////////////////////////////////////////////////////////////////////////////
class DATAPROCESS_API CVIVector3
{
private:
    double m_d[3];
    
protected:
    inline void Copy(const CVIVector3& v);
    
public:
    //-----------------------------------------------------------------------
    inline CVIVector3()                               {                          }
    inline CVIVector3(double x, double y, double z)   { Set(x, y, z);            }
    inline CVIVector3(const CVIVector3& v)            { Copy(v);                 }
    inline CVIVector3(const CVIVector2& v, double z)  { Set(v[0], v[1], z);      }
    inline CVIVector3(const double *v)                { Set(v[0], v[1], v[2]);   }
    
    //-----------------------------------------------------------------------
    inline double& operator[](int i)       { assert(i>=0 && i<3); return m_d[i]; }
    inline double  operator[](int i) const { assert(i>=0 && i<3); return m_d[i]; }
    
    inline double* Raw()                { return m_d;                       }
    inline const double* Raw() const    { return m_d;                       }
    
    //-----------------------------------------------------------------------
    inline BOOL operator==(const CVIVector3& v) const;
    inline BOOL operator!=(const CVIVector3& v) const;
    
    //-----------------------------------------------------------------------
    inline void Set(double x, double y, double z);
    inline void Set(const double *v);
    inline CVIVector3& operator=(const CVIVector3& v);
    inline CVIVector3& operator+=(const CVIVector3& v);
    inline CVIVector3& operator-=(const CVIVector3& v);
    inline CVIVector3& operator*=(double s);
    inline CVIVector3& operator/=(double s);
    inline CVIVector3& operator^=(const CVIVector3& v);
    
    //-----------------------------------------------------------------------
    inline CVIVector3 operator+(const CVIVector3& v) const;
    inline CVIVector3 operator-(const CVIVector3& v) const;
    inline CVIVector3 operator-() const;
    inline CVIVector3 operator*(double s) const;
    inline CVIVector3 operator/(double s) const;
    inline double   operator*(const CVIVector3& v) const;
    inline CVIVector3 operator^(const CVIVector3& v) const;
};

/////////////////////////////////////////////////////////////////////////////
/*-------------------------------------------------------------------------*/
inline DATAPROCESS_API double Norm(const CVIVector3& v)
{
    return sqrt(v[0]*v[0] + v[1]*v[1] + v[2]*v[2]);
}

/*-------------------------------------------------------------------------*/
inline DATAPROCESS_API double Norm2(const CVIVector3& v)
{
    return v[0]*v[0] + v[1]*v[1] + v[2]*v[2];
}

/*-------------------------------------------------------------------------*/
inline DATAPROCESS_API double Length(const CVIVector3& v)
{
    return Norm(v);
}

/*-------------------------------------------------------------------------*/
inline DATAPROCESS_API double Length2(const CVIVector3& v)
{
    return Norm2(v);
}

/*-------------------------------------------------------------------------*/
inline DATAPROCESS_API double Unitize(CVIVector3& v)
{
    double l=Norm2(v);
    if( l!=1.0 && l!=0.0 )
    {
        l = sqrt(l);
        v /= l;
    }
    return l;
}

/*-------------------------------------------------------------------------*/
inline CVIVector3 operator*(double s, const CVIVector3& v)
{
    return v*s;
}

/////////////////////////////////////////////////////////////////////////////
/*-------------------------------------------------------------------------*/
inline void CVIVector3::Copy(const CVIVector3& v)
{
    m_d[0] = v[0];  m_d[1] = v[1];  m_d[2] = v[2];
}

/*-------------------------------------------------------------------------*/
inline void CVIVector3::Set(double x, double y, double z)
{
    m_d[0] = x;     m_d[1] = y;     m_d[2] = z;
}

/*-------------------------------------------------------------------------*/
inline void CVIVector3::Set(const double *v)
{
    m_d[0] = v[0];  m_d[1] = v[1];  m_d[2] = v[2];
}

/*-------------------------------------------------------------------------*/
inline BOOL CVIVector3::operator==(const CVIVector3& v) const
{
    double dx = m_d[0] - v[0];
    double dy = m_d[1] - v[1];
    double dz = m_d[2] - v[2];
    return (dx*dx + dy*dy + dz*dz) <= FEQ_EPS2;
}

/*-------------------------------------------------------------------------*/
inline BOOL CVIVector3::operator!=(const CVIVector3& v) const
{
    double dx = m_d[0] - v[0];
    double dy = m_d[1] - v[1];
    double dz = m_d[2] - v[2];
    return (dx*dx + dy*dy + dz*dz) > FEQ_EPS2;
}

/*-------------------------------------------------------------------------*/
inline CVIVector3& CVIVector3::operator=(const CVIVector3& v)
{
    Copy(v);
    return *this;
}

/*-------------------------------------------------------------------------*/
inline CVIVector3& CVIVector3::operator+=(const CVIVector3& v)
{
    m_d[0] += v[0]; m_d[1] += v[1]; m_d[2] += v[2];
    return *this;
}

/*-------------------------------------------------------------------------*/
inline CVIVector3& CVIVector3::operator-=(const CVIVector3& v)
{
    m_d[0] -= v[0]; m_d[1] -= v[1]; m_d[2] -= v[2];
    return *this;
}

/*-------------------------------------------------------------------------*/
inline CVIVector3& CVIVector3::operator*=(double s)
{
    m_d[0] *= s;    m_d[1] *= s;    m_d[2] *= s;
    return *this;
}

/*-------------------------------------------------------------------------*/
inline CVIVector3& CVIVector3::operator/=(double s)
{
    double t = 1/s;
    m_d[0] *= t;    m_d[1] *= t;    m_d[2] *= t;
    return *this;
}

/*-------------------------------------------------------------------------*/
inline CVIVector3& CVIVector3::operator^=(const CVIVector3& v)
{
    double d0 = m_d[0]; double d1 = m_d[1]; double d2 = m_d[2];
    m_d[0] = d1*v[2] - v[1]*d2;
    m_d[1] = d2*v[0] - v[2]*d0;
    m_d[2] = d0*v[1] - v[0]*d1;
    return *this;
}

/*-------------------------------------------------------------------------*/
inline CVIVector3 CVIVector3::operator+(const CVIVector3& v) const
{
    return CVIVector3(m_d[0]+v[0], m_d[1]+v[1], m_d[2]+v[2]);
}

/*-------------------------------------------------------------------------*/
inline CVIVector3 CVIVector3::operator-(const CVIVector3& v) const
{
    return CVIVector3(m_d[0]-v[0], m_d[1]-v[1], m_d[2]-v[2]);
}

/*-------------------------------------------------------------------------*/
inline CVIVector3 CVIVector3::operator-() const
{
    return CVIVector3(-m_d[0], -m_d[1], -m_d[2]);
}

/*-------------------------------------------------------------------------*/
inline CVIVector3 CVIVector3::operator*(double s) const
{
    return CVIVector3(m_d[0]*s, m_d[1]*s, m_d[2]*s);
}

/*-------------------------------------------------------------------------*/
inline CVIVector3 CVIVector3::operator/(double s) const
{
    double t = 1/s;
    return CVIVector3(m_d[0]*t, m_d[1]*t, m_d[2]*t);
}

/*-------------------------------------------------------------------------*/
inline double CVIVector3::operator*(const CVIVector3& v) const
{
    return m_d[0]*v[0] + m_d[1]*v[1] + m_d[2]*v[2];
}

/*-------------------------------------------------------------------------*/
inline CVIVector3 CVIVector3::operator^(const CVIVector3& v) const
{
    double dx =  m_d[1]*v[2] - v[1]*m_d[2];
    double dy =  m_d[2]*v[0] - v[2]*m_d[0];
    double dz =  m_d[0]*v[1] - v[0]*m_d[1];
    return CVIVector3(dx, dy, dz);
}

///////////////////////////////////////////////////////////////////////////////
class DATAPROCESS_API CVIVector4
{
private:
    double m_d[4];
    
protected:
    inline void Copy(const CVIVector4& v);
    
public:
    //-----------------------------------------------------------------------
    inline CVIVector4()                             {                              }
    inline CVIVector4(double x, double y, double z, double w) { Set(x, y, z, w);   }
    inline CVIVector4(const CVIVector4& v)            { Copy(v);                     }
    inline CVIVector4(const CVIVector3& v, double w)  { Set(v[0], v[1], v[2], w);    }
    inline CVIVector4(const double *v)              { Set(v[0], v[1], v[2], v[3]); }
    
    //-----------------------------------------------------------------------
    inline double& operator[](int i)       { assert(i>=0 && i<4); return m_d[i]; }
    inline double  operator[](int i) const { assert(i>=0 && i<4); return m_d[i]; }
    
    inline double* Raw()                { return m_d;                       }
    inline const double* Raw() const    { return m_d;                       }
    
    //-----------------------------------------------------------------------
    inline BOOL operator==(const CVIVector4& v) const;
    inline BOOL operator!=(const CVIVector4& v) const;
    
    //-----------------------------------------------------------------------
    inline void Set(double x, double y, double z, double w);
    inline void Set(const double *v);
    inline CVIVector4& operator=(const CVIVector4& v);
    inline CVIVector4& operator+=(const CVIVector4& v);
    inline CVIVector4& operator-=(const CVIVector4& v);
    inline CVIVector4& operator*=(double s);
    inline CVIVector4& operator/=(double s);
    
    //-----------------------------------------------------------------------
    inline CVIVector4 operator+(const CVIVector4& v) const;
    inline CVIVector4 operator-(const CVIVector4& v) const;
    inline CVIVector4 operator-() const;
    inline CVIVector4 operator*(double s) const;
    inline CVIVector4 operator/(double s) const;
    inline double   operator*(const CVIVector4& v) const;
};

/////////////////////////////////////////////////////////////////////////////
/*-------------------------------------------------------------------------*/
inline DATAPROCESS_API double Norm(const CVIVector4& v)
{
    return sqrt(v[0]*v[0] + v[1]*v[1] + v[2]*v[2] + v[3]*v[3]);
}

/*-------------------------------------------------------------------------*/
inline DATAPROCESS_API double Norm2(const CVIVector4& v)
{
    return v[0]*v[0] + v[1]*v[1] + v[2]*v[2] + v[3]*v[3];
}

/*-------------------------------------------------------------------------*/
inline DATAPROCESS_API double Length(const CVIVector4& v)
{
    return Norm(v);
}

/*-------------------------------------------------------------------------*/
inline DATAPROCESS_API double Length2(const CVIVector4& v)
{
    return Norm2(v);
}

/*-------------------------------------------------------------------------*/
inline DATAPROCESS_API double Unitize(CVIVector4& v)
{
    double l=Norm2(v);
    if( l!=1.0 && l!=0.0 )
    {
        l = sqrt(l);
        v /= l;
    }
    return l;
}

/*-------------------------------------------------------------------------*/
inline CVIVector4 operator*(double s, const CVIVector4& v)
{
    return v*s;
}

/*-------------------------------------------------------------------------*/
inline CVIVector4 Cross(const CVIVector4& a, const CVIVector4& b, const CVIVector4& c)
{
    CVIVector4 result;

    double d1 = (b[2] * c[3]) - (b[3] * c[2]);
    double d2 = (b[1] * c[3]) - (b[3] * c[1]);
    double d3 = (b[1] * c[2]) - (b[2] * c[1]);
    double d4 = (b[0] * c[3]) - (b[3] * c[0]);
    double d5 = (b[0] * c[2]) - (b[2] * c[0]);
    double d6 = (b[0] * c[1]) - (b[1] * c[0]);

    result[0] = - a[1] * d1 + a[2] * d2 - a[3] * d3;
    result[1] =   a[0] * d1 - a[2] * d4 + a[3] * d5;
    result[2] = - a[0] * d2 + a[1] * d4 - a[3] * d6;
    result[3] =   a[0] * d3 - a[1] * d5 + a[2] * d6;

    return result;
}

/////////////////////////////////////////////////////////////////////////////
/*-------------------------------------------------------------------------*/
inline void CVIVector4::Copy(const CVIVector4& v)
{
    m_d[0] = v[0];  m_d[1] = v[1];  m_d[2] = v[2];  m_d[3] = v[3];
}

/*-------------------------------------------------------------------------*/
inline void CVIVector4::Set(double x, double y, double z, double w)
{
    m_d[0] = x;     m_d[1] = y;     m_d[2] = z;     m_d[3] = w;
}

/*-------------------------------------------------------------------------*/
inline void CVIVector4::Set(const double *v)
{
    m_d[0] = v[0];  m_d[1] = v[1];  m_d[2] = v[2];  m_d[3] = v[3];
}

/*-------------------------------------------------------------------------*/
inline BOOL CVIVector4::operator==(const CVIVector4& v) const
{
    double dx = m_d[0] - v[0];
    double dy = m_d[1] - v[1];
    double dz = m_d[2] - v[2];
    double dw = m_d[3] - v[3];
    return (dx*dx + dy*dy + dz*dz + dw*dw) <= FEQ_EPS2;
}

/*-------------------------------------------------------------------------*/
inline BOOL CVIVector4::operator!=(const CVIVector4& v) const
{
    double dx = m_d[0] - v[0];
    double dy = m_d[1] - v[1];
    double dz = m_d[2] - v[2];
    double dw = m_d[3] - v[3];
    return (dx*dx + dy*dy + dz*dz + dw*dw) > FEQ_EPS2;
}

/*-------------------------------------------------------------------------*/
inline CVIVector4& CVIVector4::operator=(const CVIVector4& v)
{
    Copy(v);
    return *this;
}

/*-------------------------------------------------------------------------*/
inline CVIVector4& CVIVector4::operator+=(const CVIVector4& v)
{
    m_d[0] += v[0]; m_d[1] += v[1]; m_d[2] += v[2]; m_d[3] += v[3];
    return *this;
}

/*-------------------------------------------------------------------------*/
inline CVIVector4& CVIVector4::operator-=(const CVIVector4& v)
{
    m_d[0] -= v[0]; m_d[1] -= v[1]; m_d[2] -= v[2]; m_d[3] -= v[3];
    return *this;
}

/*-------------------------------------------------------------------------*/
inline CVIVector4& CVIVector4::operator*=(double s)
{
    m_d[0] *= s;    m_d[1] *= s;    m_d[2] *= s;    m_d[3] *= s;
    return *this;
}

/*-------------------------------------------------------------------------*/
inline CVIVector4& CVIVector4::operator/=(double s)
{
    double t = 1/s;
    m_d[0] *= t;    m_d[1] *= t;    m_d[2] *= t;    m_d[3] *= t;
    return *this;
}

/*-------------------------------------------------------------------------*/
inline CVIVector4 CVIVector4::operator+(const CVIVector4& v) const
{
    return CVIVector4(m_d[0]+v[0], m_d[1]+v[1], m_d[2]+v[2], m_d[3]+v[3]);
}

/*-------------------------------------------------------------------------*/
inline CVIVector4 CVIVector4::operator-(const CVIVector4& v) const
{
    return CVIVector4(m_d[0]-v[0], m_d[1]-v[1], m_d[2]-v[2], m_d[3]-v[3]);
}

/*-------------------------------------------------------------------------*/
inline CVIVector4 CVIVector4::operator-() const
{
    return CVIVector4(-m_d[0], -m_d[1], -m_d[2], -m_d[3]);
}

/*-------------------------------------------------------------------------*/
inline CVIVector4 CVIVector4::operator*(double s) const
{
    return CVIVector4(m_d[0]*s, m_d[1]*s, m_d[2]*s, m_d[3]*s);
}

/*-------------------------------------------------------------------------*/
inline CVIVector4 CVIVector4::operator/(double s) const
{
    double t = 1/s;
    return CVIVector4(m_d[0]*t, m_d[1]*t, m_d[2]*t, m_d[3]*t);
}

/*-------------------------------------------------------------------------*/
inline double CVIVector4::operator*(const CVIVector4& v) const
{
    return m_d[0]*v[0] + m_d[1]*v[1] + m_d[2]*v[2] + m_d[3]*v[3];
}

#endif
